// CForth LowPower project ***************************************************
//
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforth.c","cforth.h",
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{

START:

    RCC->AHBENR |= RCC_AHBENR_GPIOAEN;  // Enable Port A clock
    RCC->AHBENR |= RCC_AHBENR_GPIOBEN;  // Enable Port B clock

//	initGPIO();
    GPIOA->MODER = 0x00000000;      //primarily all input
    GPIOB->MODER = 0x00000000;      //primarily all input
    GPIOA->PUPDR = 0x02A6AA8A;      //primarily all inputs pulldown 
									//except PA15,14,13,9,2
									//PA9 temporarily pullUp for AUTOEXE check
    GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown

#ifdef RS232
 	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR0);	//pull-down off, wasts current
    GPIOB->MODER &= ~(GPIO_MODER_MODER0);
    GPIOB->MODER |= GPIO_MODER_MODER0_0 ;	//set PortB0 (Comm LED) as output
    GPIOB->BSRR =  GPIO_BSRR_BS_0;			//PB0 (Comm LED) ON
#else
//pin config repeated specifically:
 	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR3);	//pull-down off, wasts current
    GPIOB->MODER &= ~(GPIO_MODER_MODER3);
    GPIOB->MODER |= GPIO_MODER_MODER3_0 ;	//set PortB3 (User LED) as output
    GPIOB->BSRR =  GPIO_BSRR_BS_3;			//PB3 (User LED) ON
#endif
/*  
//useful for debugging
	//MCO is AF0 on PortA
    GPIOA->MODER &= ~(GPIO_MODER_MODER8_0);
    GPIOA->MODER |= GPIO_MODER_MODER8_1;
    GPIOA->AFR[1] &= 0xFFFFFFF0;        //PA8 is MCO
    GPIOA->AFR[1] |= 0x00000000;
*/

	SetSysClock();

    gLEDCOUNT = 0;
    gDELAYCOUNT = 0;
    gTIMEOUT = 0;
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
	gfAuto = 0;
	gLedBusy = 2;
	gfRandom = 314159265;		//seed for xorshift32 algorithm

    SysTick_Config(SystemCoreClock / 1000);	
								//1 ms for any SystemCoreClock
								//SystemCoreClock is derived from SYSCLOCK=HSE

    msDelay(50);

//init I/O state  *************************
    op_ABORT();					//reset default Forth system data
    popFlashPage25();			//loads only gBaud, gfNumBase and gProject

    msDelay(2);
 //init communication  ********************
	initUsart2();
    msDelay(2);

    sendTB(CRR);
    sendString(gTypeStr);
    sendTB(HARDWARE_VERSION + 0x30);
    sendTB('.');
    sendNS32(REVNUM);
    sendString(gCopyStr); 

	popFlashProject()				;//if (gProject==0) start empty project
									//if virgin or gProject==0 emptyForth();

    while(1) {
		quit();
		if (query() == 0) break;
		if (gfTibMax == 0) continue;
		eval();
    }
    msDelay(100);
    goto START;					//force cold reset 
}

// ##########################################################################

// Misc Low Level Routines
//************************

#ifdef EXTCLOCK_8M
//always PLLCLOCK
void SetSysClock(void)  // ******************************************
// *************************************************************************
{
__IO uint32_t StartUpCounter = 0, HSEStatus = 0;

//SYSCLK, HCLK, PCLK configuration ----------------------------------------
//Enable HSE 
  RCC->CR |= ((uint32_t)RCC_CR_HSEON);

//Wait until HSE is ready -- if Time out is reached exit 
  do  {
    HSEStatus = RCC->CR & RCC_CR_HSERDY;
    StartUpCounter++;
  } while((HSEStatus == 0) && (StartUpCounter != HSE_STARTUP_TIMEOUT));

  if ((RCC->CR & RCC_CR_HSERDY) != RESET)  { //RESET=0 defined in stm32f0xx.h
//Enable Prefetch Buffer and set Flash Latency 
//for CLOCK_48M
    FLASH->ACR = FLASH_ACR_PRFTBE | FLASH_ACR_LATENCY;

	RCC->CFGR &= ~(RCC_CFGR_HPRE);		//clear all HPRE options
	RCC->CFGR |=  RCC_CFGR_HPRE_DIV1;

//PCLK = HCLK 
    RCC->CFGR |= (uint32_t)RCC_CFGR_PPRE_DIV1;

// useful for debugging
//	RCC->CFGR |=  RCC_CFGR_MCO_SYSCLK;	
//	RCC->CFGR |= RCC_CFGR_MCO_PRE_16;	//set for project to 48M/16=3MHz

	RCC->CFGR &= (uint32_t)((uint32_t)~(RCC_CFGR_PLLSRC | 
									RCC_CFGR_PLLXTPRE | RCC_CFGR_PLLMULL));

//PLL configuration = HSE with 8.0 MHz: HSE * 6 * 48MHz
    RCC->CFGR |= (uint32_t)(RCC_CFGR_PLLSRC_PREDIV1 | 
							RCC_CFGR_PLLXTPRE_PREDIV1 | RCC_CFGR_PLLMULL6);

//Enable PLL 
    RCC->CR |= RCC_CR_PLLON;

//Wait till PLL is ready
    while((RCC->CR & RCC_CR_PLLRDY) == 0) ;

//Select PLL as system clock source
    RCC->CFGR &= (uint32_t)((uint32_t)~(RCC_CFGR_SW));
    RCC->CFGR |= (uint32_t)RCC_CFGR_SW_PLL;

//Wait till PLL is used as system clock source
    while ((RCC->CFGR & (uint32_t)RCC_CFGR_SWS) != (uint32_t)RCC_CFGR_SWS_PLL) ;

	RCC->CFGR3 = RCC_CFGR3_USART2SW_0;	//SYSCLK selected for USART2
  }
}

#else

void SetSysClock(void)  // ******************************************
// *************************************************************************
{
__IO uint32_t StartUpCounter = 0, HSIStatus = 0;

//SYSCLK, HCLK, PCLK configuration ----------------------------------------
//Enable HSI 
  RCC->CR &= ~((uint32_t)RCC_CR_HSEON);			//precaution:switch OFF
  RCC->CR |= ((uint32_t)RCC_CR_HSION);			//should be active by default

//Wait until HSI is ready -- if Time out is reached exit 
  do  {
    HSIStatus = RCC->CR & RCC_CR_HSIRDY;
    StartUpCounter++;
  } while((HSIStatus == 0) && (StartUpCounter != HSE_STARTUP_TIMEOUT));

//  if ((RCC->CR & RCC_CR_HSIRDY) != RESET)  { //RESET=0 defined in stm32f0xx.h
//Enable Prefetch Buffer and set Flash Latency 
//for CLOCK_48M
    FLASH->ACR = FLASH_ACR_PRFTBE | FLASH_ACR_LATENCY;

	RCC->CFGR &= ~(RCC_CFGR_HPRE);				//clear all HPRE options
	RCC->CFGR |=  RCC_CFGR_HPRE_DIV1;

//PCLK = HCLK 
    RCC->CFGR |= (uint32_t)RCC_CFGR_PPRE_DIV1;

//useful for debugging
//	RCC->CFGR |=  RCC_CFGR_MCO_SYSCLK;	
//	RCC->CFGR |= RCC_CFGR_MCO_PRE_16;	//set for project to 48M/16=3MHz

	RCC->CFGR &= (uint32_t)((uint32_t)~(RCC_CFGR_PLLSRC | RCC_CFGR_PLLXTPRE 
								| RCC_CFGR_PLLMULL)); //select PLL input:HSI/2
//  CLOCK_48M
//PLL configuration = HSI/2 = 4.0 MHz: HSI * 12 * 48MHz
    RCC->CFGR |= (uint32_t)( RCC_CFGR_PLLMULL12);

//Enable PLL 
    RCC->CR |= RCC_CR_PLLON;

//Wait till PLL is ready
    while((RCC->CR & RCC_CR_PLLRDY) == 0) ;

//Select PLL as system clock source
    RCC->CFGR &= (uint32_t)((uint32_t)~(RCC_CFGR_SW));
    RCC->CFGR |= (uint32_t)RCC_CFGR_SW_PLL;

//Wait till PLL is used as system clock source
    while ((RCC->CFGR & (uint32_t)RCC_CFGR_SWS) != (uint32_t)RCC_CFGR_SWS_PLL) ;

	RCC->CFGR3 = RCC_CFGR3_USART2SW_0;	//SYSCLK selected for USART2
}

#endif
// ************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE* //

void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-042.c" hardware independent
// **************************************************************************
{
		if (gLedBusy > 1) {
			GPIOB->BSRR = GPIO_BSRR_BR_3;	// clear bit PB0,LED OFF
			gLEDCOUNT = 4; //ca 300 milliseconds
		}
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT)  servSystem();		//unconditionally in this case
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// **************************************************************************
{
uint32_t countdown;

	countdown = 51;
    while (countdown)   {
        countdown-- ;
    }
}


void servSystem(void)   // ***************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullTermRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//slowed down by SysTick to 50 times/second
// ***************************************************************************
{
	if (gSYSCOUNT)  return;		//countdown in SysTick_Handler()

//else if if (gSYSCOUNT==0 )50 Hz service interval:  *************************
    gSYSCOUNT = 50;         //restart 20Hz service interval

//LED control   28Dec24 LED blink behaviour changed for better user access
	if  (gLEDCOUNT)  {
		gLEDCOUNT--;
		if (gLEDCOUNT==0)  {
			if (gLedBusy>= 2)  	GPIOB->BSRR = GPIO_BSRR_BS_3;
											// Set bit PB3 (User LED) ON again
		}
	}
}


void popFlashPage25(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test;

	emptyForth();			//primarily init defaults - implies gProject = 0;

	flashPtr = (uint32_t *)(PAGE25);
	test = ~(*flashPtr);//virgin test can't be=0 if project was saved in Flash
	flashPtr++;		

	if (test == 0)  return;	//defaults already init
	else  {			//Flash is not virgin, Project seems to be stored
										//programmed first=read lower16bit
		gfNumBase =  test & 0x000000FF;
		if (gfNumBase == 0) gfNumBase = DEFAULT_NUMBASE;
		gfTempNumBase = gfNumBase;
	}
	gProject = (test >> 16 ) & 0x000000FF;		
										//if==0, "empty project" will be used
}


void pushFlashPage25(void)   //**********************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint16_t *flashPtr;

//unlock flash
#define KEY1    0x45670123				//mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB
    while ((FLASH->SR & FLASH_SR_BSY) !=0);//wait until no operation going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }
//erase Page25
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = PAGE25;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    flashPtr = (uint16_t *)PAGE25;
	FLASH->CR |= FLASH_CR_PG;

	//2 half-words for system and peripheral parameters
	//flash virgin test, too
    *flashPtr = (uint16_t)(~(gfNumBase) & 0x0000FFFF);  
    flashPtr++;
										//program first=read as lower16bit
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
//02Jan25 gBaud cancelled
	*flashPtr = (uint16_t)(~(gProject) & 0x0000FFFF); 
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}


void popFlashProject(void)   //**********************************************
// **************************************************************************
{
uint32_t k, n;
uint32_t  *flashPtr;
uint32_t *sramPtr;

//emptyForth() already updated in popFlashPage25() (default system parameters)

	if (gProject == 1) flashPtr = (uint32_t *)PAGE26;
	else flashPtr = (uint32_t *)PAGE29;
//4 words=16bytes for system and peripherial parameters
	k = ~(*flashPtr);			//DUMMY cannot be==0 if project saved in Flash
	flashPtr++;
	n =  ~(*flashPtr);			//cannot be==0 if project saved in Flash
	flashPtr++;
	if ((n | k) == 0) {					//project Flash is virgin
		emptyForth();						//implies gProject=0
//		return;
	}
	else  {
//from here only pre-saved project#1 or #2 are handled
		if (n & 0x00000001)  gfNumBase = 16;
		else gfNumBase = 10;
		gfTempNumBase = gfNumBase;

		gfSpiMode = (n >> 9) & 0x00000003;	
		gfSpiBits = ((n >> 11) & 0x0000001F) + 4;	//range 4...32
		gfSpiClock = (n >> 16) & 0x0000FFFF;

		k = ~(*flashPtr);
		flashPtr++;
		gfAuto = (k & 0x0000FFFF);
		gfFreq = (k >> 16) & 0x0000FFFF;

		k = ~(*flashPtr);
		flashPtr++;
		n = k & 0x0000FFFF;
		if (n > 2048) n |= 0xFFFF0000;			//manual cast to int32_t
		gfPWM1 = (int32_t)n;
		if (gfFreq) {
			subFREQ();
			subPWM1();
			n = (k >> 16) & 0x0000FFFF;
			if (n > 2048) n |= 0xFFFF0000;		//manual cast to int32_t
			gfPWM2 = (int32_t)n;
			subPWM2();
		}

//loading is possible over page limits.
//gfUserHeader[] and gfTcode[] occupy a coherent data space in flash
		sramPtr = (uint32_t *) &(gfUserHeader[0]);
		for (k = 0;k < (MAXUSER * 4);k++)  {	
									//380 words=1520 bytes=ca.1.5pages
			*sramPtr = ~(*flashPtr);
			sramPtr++;
			flashPtr++;
		}
		sramPtr = (uint32_t *) &(gfTCode[0]);
		for (k = 0;k < 0x180;k++)  {	//1536(0x600) bytes = 1.5 pages
			*sramPtr = ~(*flashPtr);
			sramPtr++;
			flashPtr++;
		}
	}
	wait5us();
	op_ABORT();					//reset project Forth system data

	//emergency brake to avoid AUTO deadlock
	if ((GPIOA->IDR & 0x0200) == 0) gfAuto = 0;	//if PA9 (RX input)is low
	GPIOA->PUPDR &= ~GPIO_PUPDR_PUPDR9;			//modification 30Jan25
	GPIOA->PUPDR |= GPIO_PUPDR_PUPDR9_1;		//pull-down = default state

	sendTB(CRR);
	op_QUEST();
	if (gfAuto) {
		gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
		interpretUserToken();
	}
}


void pushFlashProject(void)   //*********************************************
// **************************************************************************
{
uint32_t k;
uint16_t  *flashPtr;
uint32_t *sramPtr;

	if ((gProject != 1) && (gProject != 2)) return;	//precaution

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB
    while ((FLASH->SR & FLASH_SR_BSY) !=0);
								//wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page26 (gProject==1) or Page29 (gProject==2)
    FLASH->CR |= FLASH_CR_PER;
    if (gProject == 1) FLASH->AR = PAGE26;
	else FLASH->AR = PAGE29;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
//write system parameters and gfUserHeader[] into page29
//programming Flash is possible only by 16bit half-words
//first programmed 16bit half-word will be read as lower 16bits
//                                         when data are read as 32bit words!!
//second programmed 16bit half-word will be read as higher 16bits
//                                         when data are read as 32bit words!!

	if (gProject == 1) flashPtr = (uint16_t *)PAGE26;
	else flashPtr = (uint16_t *)PAGE29;

    //4 words=8 half-words=16bytes for system and peripheral parameters
    *flashPtr = (uint16_t)(~(2) & 0x0000FFFF);//DUMMY read = lower 16bit
//    *flashPtr = 0;//read = lower 16bit
							//cannot be zero, used for Flash virgin test
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;
    *flashPtr = 0;//DUMMY read = upper 16bit
							//cannot be zero, used for Flash virgin test
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;

	if (gfNumBase == 10) k = 0;
    else k = 1;
	k |= ((gBaud & 0x0000007F) << 1); // | ((gRsTerm & 1) << 8);
	k |= ((gfSpiMode & 0x0000003)<< 9) ; 
    k |= (((gfSpiBits - 4) & 0x0000001F) << 11); //burn range = 0...28
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program first=read lower16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr++;
	k =  gfSpiClock;
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program last=read upper16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;	

    *flashPtr = (uint16_t)(~gfAuto & 0x0000FFFF);	//read as lower 16bits
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;
   *flashPtr = (uint16_t)(~gfFreq & 0x0000FFFF);	//read as upper 16bits
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;

	k = (uint32_t) gfPWM1 ;
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program first=read lower16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;
	k = (uint32_t) gfPWM2 ;
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program last=read upper16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;

//gfUserHeader[] and gfTcode[] occupy a coherent data space in flash
    sramPtr = (uint32_t *) &(gfUserHeader[0]);//here:252words=1008bytes+16=1024
    for (k = 0;k < 252;k++)  { //each entry 16bytes ->first 63 entries
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)((~(*sramPtr) >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
    FLASH->CR &= ~FLASH_CR_PG;

//erase Page27 (gProject==1) or Page30 (gProject==2)
    FLASH->CR |= FLASH_CR_PER;
	if (gProject == 1) FLASH->AR = PAGE27;
	else FLASH->AR = PAGE30;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
	if (gProject == 1) flashPtr = (uint16_t *)PAGE27;
	else flashPtr = (uint16_t *)PAGE30;
//write remaining part of gfUserHeader:MAXUSER=(95-63) *4=128 words total
    for (k = 0;k < 128;k++)  { //each entry 16bytes --> last 32entries
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)((~(*sramPtr) >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
//flashed up to 0x8007A00
//write first part of gfTCode into page 27(gProject==1) or page30(gProject==2)
    sramPtr = (uint32_t *) &(gfTCode[0]);	//WORD addressed!
    for (k = 0;k < 0x80;k++)  {				//128words=512bytes=0.5 page
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)(~(*sramPtr >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
    FLASH->CR &= ~FLASH_CR_PG;

//erase Page28 (gProject==1) or Page31 (gProject==2)
    FLASH->CR |= FLASH_CR_PER;
	if (gProject == 1) FLASH->AR = PAGE28;
	else FLASH->AR = PAGE31;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
//write second half of gfTCode into  28(gProject==1) or page31(gProject==2)
	if (gProject == 1) flashPtr = (uint16_t *)PAGE28;
	else flashPtr = (uint16_t *)PAGE31;
    sramPtr = (uint32_t *)gfTCode + 0x80;	//start addr = 0x200bytes
    for (k = 0;k < 0x100;k++)  { //0x400bytes= 1 page, word read,halfWord burn
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)(~(*sramPtr >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
//flash is 100% filled
//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}

//###########################################################################

//USART Support
//*************
void initUsart2(void)  // ****************************************************
// if USB is used with TSSOP package, USART2 is the only available USART
// USART2-TxD is PortA,2
// USART2-RxD is PortA,15
// **************************************************************************
{
uint16_t baudset;

    RCC->AHBENR |= RCC_AHBENR_GPIOAEN ;     // PortA clock enable
    RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ; // USART2 clock enable

    GPIOA->MODER &= ~(GPIO_MODER_MODER2_0 | //Set pin 2,3 to alternate function
                      GPIO_MODER_MODER15_0 ) ;  // 10=Alternate function mode
    GPIOA->MODER |= (GPIO_MODER_MODER2_1 |  //Set pin 2,3 to alternate function
                     GPIO_MODER_MODER15_1 ) ;  // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

//USART2 is AF1
    GPIOA->AFR[0] = 0x00000100;
    GPIOA->AFR[1] = 0x10000000;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

	USART2->BRR =  0x01A0;		//115200Baud: 26.0 at 48 MHz APB1 Clock;
    USART2->CR2 = 0;				//default:1 Stopbit
    USART2->CR3 = 0;				//default:no Handshake
//the TXIE flag will be set when the first byte is going to be transmitted
    USART2->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |
                  USART_CR1_TE | USART_CR1_RE;  //8Bit, no Parity,RX&TX Int ON
    NVIC_EnableIRQ(USART2_IRQn);
}


void sendTB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR & USART_ISR_TXE));
    USART2->TDR = data;
	if (gLedBusy == 3) LED_Off();
}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
	sendTB(asciiNibble);
}


void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data & 0xF0) >> 4;
    temp = temp  +0x30;
    if (temp > 0x39) temp+= 7;
    sendTB(temp);
    temp= data & 0x0F;
    temp = temp + 0x30;
    if (temp > 0x39) temp += 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// Leading zeroes are suppressed
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(0x20);

}


void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format WITH leading 0x.Leading zeroes suppressed
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


uint32_t pullTermRxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART receive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();		//serve watchdog etc.

	if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
		recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
		if (gUSART_RXTAIL==gUSART_RXBUFTOP)  gUSART_RXTAIL=gUSART_RXBUF;
		LED_Off();
		return recv;
	}
	else return 0;
}

// Interrupt Handlers
//*******************

void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if (gSYSCOUNT) gSYSCOUNT--;			//controls servSystem()
	if (gDELAYCOUNT) gDELAYCOUNT--;
	if (gTIMEOUT) gTIMEOUT--;			//to be used in any subroutine
	if (gfMSCOUNT) gfMSCOUNT --;		//used by MS, DS
	if (gfTIXCOUNT) gfTIXCOUNT --;		//used by  TIX, WAIT
	if (gfSTEPCOUNT > 0) gfSTEPCOUNT--;	//stepper timing
	gfRandom ^= gfRandom << 13;			//xorshift32 algorithm/white noise
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;
}


void USART2_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t test;

	if (USART2->ISR & USART_ISR_RXNE)  {
		test = USART2->RDR & 0x00FF;
		if (test == CTRL_C)  {
			gfTermInput = CTRL_C;
			return;
		}
		if (test != CTRL_B)  {
			*gUSART_RXHEAD = (uint8_t)test;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
        else  op_BAKOP();
	}
}

// #########################################################################
